"""
Utilities for waiting.
"""
import asyncio
import logging
import time

from openoligo.utils.sim import SIMULATION_SPEEDUP_FACTOR


async def wait_async(seconds: float) -> None:
    """Wait for a given number of seconds."""
    logging.debug("Start waiting for %.2f seconds", seconds)  # pragma: no cover
    await asyncio.sleep(seconds / SIMULATION_SPEEDUP_FACTOR)
    logging.debug("Done waiting for %.2f seconds", seconds)  # pragma: no cover


def wait(seconds: float) -> None:
    """Wait for a given number of seconds."""
    logging.debug("Start waiting for %.2f seconds", seconds)
    time.sleep(seconds / SIMULATION_SPEEDUP_FACTOR)
    logging.debug("Done waiting for %.2f seconds", seconds)


def ms(seconds: float) -> float:  # pylint: disable=invalid-name
    """Convert seconds to milliseconds."""
    return seconds / 1000



"""
Sequence class for DNA, RNA and modified Nucleic Acid sequences.
"""
import re


def is_valid_dna(sequence):
    """
    Check if a sequence is a valid DNA sequence.
    """
    pattern = r"^[ATCGatcg]*$"
    if re.match(pattern, sequence):
        return True
    return False


class Seq:
    """
    Representation of the sequence as a string.
    """

    def __init__(self, seq: str) -> None:
        """
        Initialize the sequence.
        """
        self.seq = seq.upper()
        if not is_valid_dna(self.seq):
            raise ValueError("Invalid DNA sequence")
        self.index = 0

    def __iter__(self):
        """
        Allows iteration over the sequence.
        """
        return self

    def __next__(self) -> str:
        """
        Returns the next base in the sequence.

        raises:
            StopIteration: if the end of the sequence is reached.
        """
        if self.index < len(self.seq):
            result = self.seq[self.index]
            self.index += 1
            return result
        raise StopIteration

    def __repr__(self) -> str:
        """
        Returns string representation of the sequence.
        """
        return self.seq

    def __len__(self) -> int:
        """
        get the length of the sequence.
        """
        return len(self.seq)

    def __getitem__(self, key) -> str:
        """
        Allows indexing of the sequence.
        """
        return self.seq[key]



"""
Imports all the modules in the utils directory.
"""
from openoligo.utils.sim import SIMULATION_SPEEDUP_FACTOR
from openoligo.utils.wait import ms, wait, wait_async

__all__ = ["ms", "wait", "wait_async", "SIMULATION_SPEEDUP_FACTOR"]



"""
This module contains definitions of protocols and exceptions related to switchable devices.
"""
from enum import Enum
from typing import Iterator, Protocol

from openoligo.hal.platform import (
    PLATFORM_TO_BOARD,
    MinimumCommonPinout,
    Platform,
    __platform__,
    rpi_board_pins,
)


class Board:
    """
    Represents a board with a minimum common pinout.
    """

    board_dict: MinimumCommonPinout

    def __init__(self, platform: Platform) -> None:
        """Creates a board given a particular platform."""
        self.board_dict = PLATFORM_TO_BOARD.get(platform, rpi_board_pins)

    def __getattr__(self, name: str) -> str:
        """Return the pin number (as a string) for a given pin name."""
        if name.startswith("P"):
            try:
                return self.board_dict[name]
            except KeyError as exc:
                raise AttributeError(f"{self.__class__.__name__} has no pin {name}") from exc
        raise AttributeError(f"{self.__class__.__name__} has no attribute {name}")

    def __len__(self) -> int:
        """Return the number of pins on the board."""
        return len(self.board_dict)

    def __repr__(self) -> str:
        """Return a string representation of the board."""
        return f"{self.__class__.__name__}({self.board_dict})"

    def __iter__(self) -> Iterator[tuple[str, str]]:
        """Return an iterator over the pin names."""
        return iter(self.board_dict.items())


board = Board(__platform__)


class GpioMode(Enum):
    """GPIO mode."""

    IN = 0
    OUT = 1


class ValveState(Enum):
    """This class represents the state of a valve."""

    OPEN_FLOW = 0
    CLOSED_FLOW = 1


class ValveType(Enum):
    """Valve Type: NC/NO"""

    NORMALLY_CLOSED = 0
    NORMALLY_OPEN = 1


class ValveRole(Enum):
    """Valve Role: Inlet/Outlet/Transit/Branch"""

    INLET = 0
    OUTLET = 1
    TRANSIT = 2
    BRANCH = 3


class Switchable(Protocol):
    """
    This protocol represents a switchable device. Any class implementing this protocol should
    be able to set, get the current value and toggle the state of the switch.
    """

    def __init__(self, pin: int, name: str):
        """Initialize the switchable device."""
        raise NotImplementedError

    def set(self, state: bool):
        """Set the state of the switch."""
        raise NotImplementedError

    #    @property
    #    def gpio(self) -> Board:
    #        """Get the GPIO pin number."""
    #        raise NotImplementedError

    @property
    def value(self) -> bool:
        """Get the current value of the switch."""
        raise NotImplementedError


class Valvable(Switchable, Protocol):
    """
    A semantic wrapper around a switchable device. This class is used to represent a valve.
    """

    def __init__(self, pin: int, name: str, valve_type: ValveType = ValveType.NORMALLY_OPEN):
        """Initialize the valve."""
        raise NotImplementedError

    def open(self):
        """Open the valve."""
        raise NotImplementedError

    def close(self):
        """Close the valve."""
        raise NotImplementedError

    @property
    def get_type(self) -> ValveType:
        """Get the type of the valve."""
        raise NotImplementedError


class SwitchingError(Exception):
    """
    An exception that represents a failure in switching operation.
    """


class InvalidManifoldSizeError(Exception):
    """
    An exception that represents an invalid manifold size.
    """


class NoSuchPinInPinout(Exception):
    """
    Thrown when a pin for a given name is not found in the pinout.
    """


class OneSourceException(SwitchingError):
    """
    When more than one source valve is being attempted to be used.
    """


class OneDestinationException(SwitchingError):
    """
    When more than one destination valve is being attempted to be used.
    """



"""
GPIO access abstraction: RPi.GPIO on Raspberry Pi, MockGPIO otherwise.
"""
import importlib
import logging
from abc import ABC, abstractmethod

from openoligo.hal.platform import Platform, __platform__
from openoligo.hal.types import GpioMode, board


def get_gpio():
    """Get the GPIO module."""
    if __platform__ == Platform.BB:
        return MockGPIO()
    if __platform__ == Platform.RPI:
        gpio = importlib.import_module("RPi.GPIO")
        return RPiGPIO(gpio)
    return MockGPIO()


class GPIOInterface(ABC):
    """GPIO interface."""

    @abstractmethod
    def setup_pin(self, pin: str, mode: GpioMode) -> None:
        """Set up a GPIO pin."""

    @abstractmethod
    def set(self, pin: str, value: bool) -> None:
        """Set the output value of a GPIO pin."""
        logging.debug("Setting pin %s to %s", pin, value)

    @abstractmethod
    def value(self, pin: str) -> bool:
        """Get the value of a GPIO pin."""

    @abstractmethod
    def cleanup(self) -> None:
        """Clean up GPIO."""

    @abstractmethod
    def __repr__(self) -> str:
        """Return a string representation of the GPIO module."""


class RPiGPIO(GPIOInterface):
    """GPIO access on Raspberry Pi."""

    def __init__(self, gpio):
        """Import RPi.GPIO"""
        self.gpio = gpio
        self.gpio.setmode(self.gpio.BCM)
        self.gpio.setwarnings(False)

    def setup_pin(self, pin: str, mode: GpioMode = GpioMode.OUT) -> None:
        """Set up a GPIO pin."""
        self.gpio.setup(pin, self.gpio.OUT if mode == GpioMode.OUT else self.gpio.IN)

    def set(self, pin: str, value: bool) -> None:
        """Set the output value of a GPIO pin."""
        super().set(pin, value)
        self.gpio.output(int(pin), self.gpio.HIGH if value else self.gpio.LOW)

    def value(self, pin: str) -> bool:
        return self.gpio.input(pin)

    def cleanup(self):
        """Clean up GPIO."""
        self.gpio.cleanup()

    def __repr__(self) -> str:
        """Return a string representation of the GPIO module."""
        return ", ".join(f"({pin}, {1 if self.value(pin) else 0})" for pin, _ in board)


class MockGPIO(GPIOInterface):
    """Mock GPIO access."""

    def __init__(self):
        """
        Initialize the state. Its keys are the pin numbers,
        and the values are their mock states
        """
        self.state: dict[str, bool] = {}
        for pin, _ in board:
            self.state[pin] = False

    def setup_pin(
        self, pin: str, _: GpioMode = GpioMode.OUT
    ) -> None:  # pylint: disable=unused-argument
        """Set up a GPIO pin."""
        self.state[pin] = False

    def set(self, pin: str, value: bool) -> None:
        """Set the output value of a GPIO pin."""
        # super().set(pin, value)
        if pin.startswith("P"):
            self.state[pin] = value
        else:
            self.state[f"P{pin}"] = value

    def value(self, pin: str) -> bool:
        """Get the value of a GPIO pin."""
        if pin.startswith("P"):
            return self.state[pin]
        return self.state[f"P{pin}"]

    def cleanup(self):
        """Clean up GPIO."""
        for pin, _ in board:
            self.set(pin, False)

    def __repr__(self) -> str:
        """Return a string representation of the GPIO module."""
        # return ", ".join(f"({pin}, {self.value(pin)})" for pin, _ in board)
        return self.state.__repr__()



"""
Provides the platform on which the program is running.
"""
from enum import Enum


def is_rpi() -> bool:
    """Returns True if running on a Raspberry Pi, False otherwise."""
    try:
        with open("/proc/cpuinfo", "r", encoding="utf-8") as file:
            return "Raspberry" in file.read()
    except FileNotFoundError:
        return False


def is_bb() -> bool:
    """Returns True if running on a Raspberry Pi, False otherwise."""
    try:
        with open("/proc/cpuinfo", "r", encoding="utf-8") as file:
            return "AM33XX" in file.read()
    except FileNotFoundError:
        return False


class Platform(Enum):
    """Platform type."""

    RPI = "RPI"
    BB = "BB"
    SIM = "SIM"


def get_platform() -> Platform:
    """Get the platform type."""
    if is_rpi():
        return Platform.RPI
    if is_bb():
        return Platform.BB
    return Platform.SIM


__platform__ = get_platform()


MinimumCommonPinout = dict[str, str]


rpi_board_pins: MinimumCommonPinout = {
    "P3": "3",
    "P5": "5",
    "P7": "7",
    "P8": "8",
    "P10": "10",
    "P11": "11",
    "P12": "12",
    "P13": "13",
    "P15": "15",
    "P16": "16",
    "P18": "18",
    "P19": "19",
    "P21": "21",
    "P22": "22",
    "P23": "23",
    "P24": "24",
    "P26": "26",
    "P27": "27",
    "P28": "28",
    "P29": "29",
    "P31": "31",
    "P32": "32",
    "P33": "33",
    "P35": "35",
    "P36": "36",
    "P37": "37",
    "P38": "38",
    "P40": "40",
}


bb_board_pins: MinimumCommonPinout = {
    "P3": "P8_11",
    "P5": "P8_12",
    "P7": "P8_13",
    "P8": "P8_14",
    "P10": "P8_15",
    "P11": "P8_16",
    "P12": "P8_17",
    "P13": "P8_18",
    "P15": "P8_19",
    "P16": "P8_20",
    "P18": "P8_21",
    "P19": "P8_22",
    "P21": "P8_23",
    "P22": "P8_24",
    "P23": "P8_25",
    "P24": "P8_26",
    "P26": "P8_27",
    "P27": "P8_28",
    "P28": "P8_29",
    "P29": "P8_30",
    "P31": "P8_31",
    "P32": "P8_32",
    "P33": "P8_33",
    "P35": "P8_34",
    "P36": "P8_35",
    "P37": "P8_36",
    "P38": "P8_37",
    "P40": "P8_38",
}


PLATFORM_TO_BOARD: dict[Platform, MinimumCommonPinout] = {
    Platform.BB: bb_board_pins,
    Platform.RPI: rpi_board_pins,
    Platform.SIM: rpi_board_pins,
}
